import sys
import numpy as np
import scipy
from .utils import *
from mpi4py import MPI

comm = MPI.COMM_WORLD



def get_linear_indexes(mfp,value):


   n = len(mfp)

   if value < mfp[0]:
      aj = (value-mfp[0])/(mfp[1]-mfp[0]) 
      return 0,1-aj,1,aj  
   elif value > mfp[-1]:
      aj = (value-mfp[n-2])/(mfp[n-1]-mfp[n-2]) 
      return n-2,1-aj,n-1,aj  
   else:
    for m in range(n-1):
      if (value <= mfp[m+1]) and (value >= mfp[m]) :
        aj = (value-mfp[m])/(mfp[m+1]-mfp[m]) 
        return m,1-aj,m+1,aj  




def mfp3D(**argv): 

   #load data
   if argv.setdefault('read_from_file',True):
     data = load_data('mfp')
     Kacc = data['Kacc']
     mfp_bulk  = data['mfp']
   else:  
     mfp_bulk = argv['mfp']
     Kacc    = argv['Kacc']

   #Get the distribution
   kappa_bulk = Kacc[-1]
   kappa_bulk = np.zeros_like(Kacc)
   kappa_bulk[0] = Kacc[0]



   #Polar Angle-----
   n_phi = int(argv.setdefault('n_phi',48)); Dphi = 2.0*np.pi/n_phi
   #phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
   phi = np.linspace(0,2.0*np.pi,n_phi,endpoint=False)
   #--------------------
   #Azimuthal Angle------------------------------
   n_theta = int( argv.setdefault('n_theta',24))
   sym = 1

   Dtheta = np.pi/n_theta
   theta = np.linspace(Dtheta/2.0,np.pi - Dtheta/2.0,n_theta)
   dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)

   #Compute directions---
   polar = np.array([np.sin(phi),np.cos(phi),np.ones(n_phi)]).T
   azimuthal = np.array([np.sin(theta),np.sin(theta),np.cos(theta)]).T
   direction = np.einsum('ij,kj->ikj',azimuthal,polar)

   ftheta = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
   fphi= np.sinc(Dphi/2.0/np.pi)
   polar_ave = np.array([fphi*np.ones(n_phi),fphi*np.ones(n_phi),np.ones(n_phi)]).T
   azimuthal_ave = np.array([ftheta,ftheta,np.cos(Dtheta/2)*np.ones(n_theta)]).T

   direction_ave = np.multiply(np.einsum('ij,kj->ikj',azimuthal_ave,polar_ave),direction)
   domega = np.outer(dtheta, Dphi * np.ones(n_phi))

   n_mfp = argv.setdefault('n_mfp',50)
   #Total numbre of momentum discretization----
   nm = n_phi * n_mfp * n_theta
   #------------------------------------------


   n_mfp_bulk = len(mfp_bulk)

   if len(mfp_bulk) == n_mfp:
     mfp_sampled = mfp_bulk
   else:  
     mfp_sampled = np.logspace(min([-2,np.log10(min(mfp_bulk)*0.99)]),np.log10(max(mfp_bulk)*1.01),n_mfp)#min MFP = 1e-2 nm 


   n_mfp = len(mfp_sampled)
   temp_coeff = np.zeros(nm) 
   kappa_directional = np.zeros((n_mfp,n_theta*n_phi,3)) 
   kdp = np.zeros((n_mfp,n_theta*n_phi,3)) 
   kd = np.zeros((n_mfp,n_theta*n_phi,3)) 

   #suppression = np.zeros((n_mfp,n_phi*n_theta,n_mfp_bulk)) 
   temp_coeff = np.zeros((n_mfp,n_theta*n_phi))
   tcp = np.zeros((n_mfp,n_theta*n_phi))
   tc = np.zeros((n_mfp,n_theta*n_phi))

   g1 = kappa_bulk/mfp_bulk/mfp_bulk*sym
   g2 = kappa_bulk/mfp_bulk*sym

   n_angles = n_theta*n_phi
   dirr = sym*np.einsum('tpi,tp->tpi',direction_ave,domega).reshape((n_angles,3))
   domega = domega.reshape((n_angles))

   block =  n_mfp_bulk//comm.size
   rr = range(block*comm.rank,n_mfp_bulk) if comm.rank == comm.size-1 else range(block*comm.rank,block*(comm.rank+1))

   mfp_sampled_log = np.log10(mfp_sampled)
   for m in rr:  

    if argv.setdefault('log_interpolation',False):     
     (m1,a1,m2,a2) = get_linear_indexes(mfp_sampled_log,np.log10(mfp_bulk[m]))
    else: 
     (m1,a1,m2,a2) = get_linear_indexes(mfp_sampled,mfp_bulk[m])
   
    kdp[m1] += a1 * g2[m]*dirr
    kdp[m2] += a2 * g2[m]*dirr
    tcp[m1] += a1 * g1[m]*domega
    tcp[m2] += a2 * g1[m]*domega

        
   comm.Allreduce([kdp,MPI.DOUBLE],[kd,MPI.DOUBLE],op=MPI.SUM)
   comm.Allreduce([tcp,MPI.DOUBLE],[tc,MPI.DOUBLE],op=MPI.SUM)

   

   #----
   kd *= 3/(4.0*np.pi)
   tc /= np.sum(tc)
   #----

   direction = direction_ave.reshape((n_theta * n_phi,3))
   F = np.einsum('m,dj->mdj',mfp_sampled,direction)

   rhs_average = mfp_sampled*mfp_sampled/3
   thermal_conductance = 3*mfp_sampled/2
   #-----------------------------

   #replicate bulk values---
   kappa = np.zeros((3,3))
   for t in range(n_theta): 
     for p in range(n_phi): 
      for m in range(n_mfp): 
        index = t * n_phi + p 
        tmp = kd[m,index]
        kappa += np.outer(tmp,direction_ave[t,p])*mfp_sampled[m]



   return {'tc':tc,\
           'VMFP':direction,\
           'sigma':kd,\
           'kappa':kappa,\
           'ang_coeff':domega.reshape(n_phi*n_theta)/4/np.pi,\
           'model':np.array([6]),\
           'mfp_average':rhs_average*1e18,\
           'thermal_conductance':thermal_conductance*1e9,\
           'mfp_sampled':mfp_sampled,\
           'suppression':np.zeros(1),\
           'kappam':np.zeros(1),\
           'mfp_bulk':mfp_bulk}





import numpy as np
from openbte.utils import *
from pubeasy import MakeFigure

def plot_mode_kappa(**argv):
 #-----------------------
 mat = load_data('material')
 data = load_data('rta')
 mfp_0 = 1e-9
 #I = np.where(np.linalg.norm(mfp_bulk,axis=1) > mfp_0)[0]
 I = np.arange(len(data['tau']))
 tau = data['tau'][I]
 v = data['v'][I]
 C = data['C'][I]
 f = data['f'][I]
 #f = np.divide(np.ones_like(tau),tau, out=np.zeros_like(tau), where=tau!=0)
 kappam = np.einsum('u,u,u,u->u',tau,C,v[:,0],v[:,0]) 
 Jc = np.einsum('k,ki->ki',C,v[:,:2])
 nm = len(C)
 mfp_bulk = np.einsum('ki,k->ki',v,tau)
 r = np.linalg.norm(mfp_bulk[:,:2],axis=1)
 phi_bulk = np.array([np.arctan2(m[0],m[1]) for m in mfp_bulk[:,:2]])
 phi_bulk[np.where(phi_bulk < 0) ] = 2*np.pi + phi_bulk[np.where(phi_bulk <0)]
 kappa = data['kappa']
 #---------------------------

 #Compute MFPs
 mat = load_data('material')
 mfp_sampled  = mat['mfp_sampled']*1e9
 [n_phi,n_theta,n_mfp]  = mat['sampling']
 vmfp  = mat['VMFP']
 sigma  = mat['sigma']
 sigma = sigma[:,:,0].T
 Dphi = 2*np.pi/n_phi
 phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 phi = list(phi)
 #--------------------------------

 mat = load_data('geometry')
 dirr = int(mat['meta'][-1])

 #--------


 #Compute kappa
 sol = load_data('solver')
 T_sampled = sol['kappa_mode']

 kappa_fourier = sol['fourier']

 #Interpolation to get kappa mode
 ratio = kappa_fourier/kappa[0,0]

 T_mode = np.zeros_like(kappam)
 for m in range(nm):

     (m1,a1,m2,a2) = interpolate(mfp_sampled,r[m]*1e9,bounds='extent')
     (p1,b1,p2,b2) = interpolate(phi,phi_bulk[m],bounds='periodic',period=2*np.pi)

     T_mode[m] += T_sampled[m1,p1]*a1*b1
     T_mode[m] += T_sampled[m1,p2]*a1*b2
     T_mode[m] += T_sampled[m2,p1]*a2*b1
     T_mode[m] += T_sampled[m2,p2]*a2*b2

 T_mode *=1e9

 kappa_nano = np.einsum('i,i,i->i',C,v[:,0],T_mode)
 kappa_bulk = np.einsum('i,i,i->i',C,v[:,0],mfp_bulk[:,0])
 kappa_nano_ori = np.einsum('ij,ji',sigma,T_sampled)*1e9

 print('Check:')
 print('kappa (bulk):  ' + str(round(np.sum(kappa_bulk),2)) + ' W/m/K')
 print('kappa (mode-sampled):  ' + str(round(np.sum(kappa_nano),2)) + ' W/m/K')
 print('kappa (mfp-sampled): ' + str(round(kappa_nano_ori,2)) + ' W/m/K')


 mfp_nano = mfp_bulk.copy()


 mfp_nano[:,dirr] = T_mode

 if argv.setdefault('show',True):
  fig = MakeFigure()
  fig.add_plot(f*1e-12,mfp_nano[:,dirr]*1e6,model='scatter',color='r')
  fig.add_plot(f*1e-12,mfp_bulk[:,dirr]*1e6,model='scatter',color='b')
  fig.add_labels('Frequency [THz]','Mean Free Path [$\mu$m]')
  fig.finalize(grid=True,yscale='log',write = True,show=True,ylim=[1e-4,1e2])

 data = {'kappa_nano':kappa_nano,'mfp_nano':mfp_nano[:,dirr],'kappa_fourier':kappa_fourier,'mfp_bulk':mfp_bulk[:,dirr],'f':f}
 if argv.setdefault('save',True):
     save_data('kappa_mode',data)

 return data


if __name__ == "__main__":

 main()



import numpy as np
from numpy import inf
import sys
from .utils import *
import sys


def main():
 data = load_data(sys.argv[1][:-4])

 kappam = np.einsum('u,u,u,u->u',data['tau'],data['C'],data['v'][:,0],data['v'][:,0]) 
 mfp_bulk = np.einsum('ki,k->ki',data['v'],data['tau'])

 mfp = np.array([np.linalg.norm(m) for m in mfp_bulk])

 I = np.argsort(mfp)

 kappa = kappam[I]
 mfp = mfp[I]
 Kacc = np.zeros_like(kappa)
 
 for n,k in enumerate(kappa):
     if n == 0: 
        Kacc[n] = k
     else:  
        Kacc[n] = k + Kacc[n-1]

 #select only meaningful MFPs
 a = np.where(Kacc < Kacc[-1]*1e-5)[0][-1]
 Kacc = Kacc[a:]
 mfp =  mfp[a:]

 #DownSamplimg
 N = 100

 mfp_sampled = np.logspace(np.log10(min(mfp)),np.log10(max(mfp)),N,endpoint=True)
 mfp_sampled[-1]= mfp[-1]
 
 
 Kacc_d = np.zeros(N)
 for n,m in enumerate(mfp_sampled):
    Kacc_d[n] = Kacc[np.where(mfp >= m)[0][0]]

 Kacc_d[-1] = Kacc[-1]
 save_data('mfp',{'Kacc':np.array(Kacc_d),'mfp':np.array(mfp_sampled)})




 



import h5py
import numpy as np
import scipy.linalg
from scipy.linalg import pinvh
import os
import hdfdict
import time
import h5py as h5
from .utils import *


def check_symmetric(a, rtol=1e-05, atol=1e-6):
    return np.allclose(a, a.T, rtol=rtol, atol=atol)

def main():

 #Get unitcell volume
 if os.path.isfile(os.sys.argv[1]) :

  cp = []
  with open(os.sys.argv[1],'r') as fh :
    dummy  = fh.readline()
    alat   = float(fh.readline())
    for i in range(3): 
     g = fh.readline().split()
     cp.append([float(g[0]),float(g[1]),float(g[2])])  
    factor = 1
    cp = np.array(cp)*alat
    V = abs(np.dot(cp[0],np.cross(cp[1],cp[2])))*1e-30

 else:
    print('No unitcell found')
    quit()

 factor = 1

 nx = int(os.sys.argv[2])
 ny = int(os.sys.argv[3])
 nz = int(os.sys.argv[4])
 T = int(os.sys.argv[5])

 cp = np.array(cp)*alat
 V = abs(np.dot(cp[0],np.cross(cp[1],cp[2])))*1e-30

 tail = str(nx) + str(ny) + str(nz) + '.hdf5'
 
 #KAPPA-------------------------
 #f = dd.io.load('kappa-m' + tail)
 
 f = hdfdict.load('kappa-m' + tail)
 
 mode_kappa = f['mode_kappa']
 #weight = f['weight'][:]
 g = f['gamma'][:]
 gg = np.pi * g[0]*1e12 #NOTE: there should be a factor 4 here according to doc.
 (nq,nb) = np.shape(g[0])
 nm = nq*nb
 alpha = V*nq
 v = np.array(f['group_velocity'])*1e2 #m/2
 w = np.array(f['frequency'])*1e12 #1/s
 q = 1.60218e-19 #C
 kb = 1.380641e-23 #j/K
 h = 6.626070151e-34 #Js
 eta = w*h/T/kb/2
 C = kb*np.power(eta,2)*np.power(np.sinh(eta),-2) #J/K


 f = gg.reshape(nm)
 I = np.where(f > 0.0)
 tau = np.zeros(nm)
 tau[I] = 1/f[I]
 w = w.reshape(nm)
 v = np.array([v[:,:,0].reshape(nb*nq),v[:,:,1].reshape(nb*nq),v[:,:,2].reshape(nb*nq)])
 v = v.T
 C = C.reshape(nm)
 ftol = 1e-30
 index = (np.logical_and(C>ftol,f>ftol)).nonzero()[0]
 exclude = (np.logical_or(C<=ftol,f<=ftol)).nonzero()[0]

 C = C[index]
 v = v[index]
 w = w[index]
 tau = tau[index]
 sigma = np.einsum('i,ij->ij',C,v)
 kappa = np.einsum('li,lj,l,l->ij',v,v,tau,C)/alpha
 print('KAPPA (RTA):')
 print(kappa)

 #---------------------------------

 #FULL MATRIX----------------------------------
 fname = 'unitary-m' + tail
 #f = dd.io.load(fname)

 f = hdfdict.load('unitary-m' + tail)
 Q = f['unitary_matrix'][0,0]
  
 #f = dd.io.load('coleigs-m' + tail)
 f = hdfdict.load('coleigs-m' + tail)
 D = f['collision_eigenvalues'][0]
 Dm = np.diag(D)
 Q = Q.reshape(nm,nm)
 
 

 #A = np.matmul(Q.T,np.matmul(Dm,Q))*1e12*np.pi
 
 QT = Q.T
 A = np.matmul(QT,(D*QT).T)
 A = np.delete(A,exclude,0)
 A = np.delete(A,exclude,1)


 #a = np.einsum('ij,i->j',A,np.sqrt(C))
 #print(sum(a))
 #print(sum(np.absolute(a)))
 #show()

 W = np.einsum('ij,i,j->ij',A,np.sqrt(C),np.sqrt(C))*1e12*np.pi
 #W = np.einsum('ij,i,j->ij',A,1/np.sqrt(C),1/np.sqrt(C))*1e12*np.pi

 print('Start inversion...')
 kappa = np.einsum('li,lk,kj->ij',sigma,pinvh(W),sigma)/alpha
 print('KAPPA (FULL):')
 print(kappa)

 data = {'W':W,'v':v,'C':C,'kappa':kappa,'alpha':np.array([alpha])}
 #np.savez_compressed('full.npz',data)   
 #Saving data
 save_data('full',data)   
 #hdfdict.dump(data,'full.h5')

 data = {'C':C/alpha,'tau':tau,'v':v,'kappa':kappa}
 saveCompressed('rta.npz',**data)   
 #save_data('full',data)   
 #np.savez_compressed('rta.npz',data)   

 #---------------------------------------------

if __name__ == '__main__':
  
  main(os.sys.argv)







from __future__ import absolute_import
import numpy as np
from scipy.sparse.linalg import splu
from termcolor import colored, cprint 
from openbte.utils import *
import scipy.sparse as sp
import time
import sys
import scipy
from cachetools import cached
from scipy.sparse.linalg import gmres
import scipy.sparse.linalg as spla
from openbte.utils import *
from openbte.fourier import *
from mpi4py import MPI

comm = MPI.COMM_WORLD


  


def sparse_dense_product(i,j,data,X):
     '''
     This solves B_ucc' X_uc' -> A_uc

     B_ucc' : sparse in cc' and dense in u. Data is its vectorized data COO descrition
     X      : dense matrix
     '''

     tmp = np.zeros_like(X)
     np.add.at(tmp.T,i,data.T * X.T[j])   

     return tmp


def solve_full(argv):

   #here you go---------------------------
   max_bte_iter   =  argv.setdefault('max_bte_iter',100)
   max_ggmres_iter   =  argv.setdefault('max_ggmres_iter',100)
   tolerance = argv.setdefault('tolerance',1e-13)
   GGMRES = argv.setdefault('ggmres',True)

   #---------------------------


   #Import data---
   mesh    = argv['geometry']
   mat     = argv['material']
   f = argv['fourier']
   factor  = mat['alpha'][0][0]
   W       = mat['W'] 
   Wdiag   = np.diag(W)
   sigma   = mat['sigma']*1e9
   Wod     = W-np.diag(np.diag(W))
   Winv    = load_data('Winv')
   WS      = load_data('winvsigma')

   gradT = compute_grad(f['temperature'],**argv)
   CT   = Wdiag/np.sum(Wdiag)

   #parallel set up------------
   n_parallel = len(sigma)
   block =  n_parallel//comm.size
   rr = range(block*comm.rank,n_parallel) if comm.rank == comm.size-1 else range(block*comm.rank,block*(comm.rank+1))
   n_block = len(rr)


   n_elems = len(mesh['elems'])
   X0 = np.tile(f['temperature'],(W.shape[0],1))
   grad_fourier = f['grad']

   #preparation
   G = np.einsum('qj,jn->qn',sigma[rr],mesh['k'],optimize=True)
   Gp = G.clip(min=0); Gm = G.clip(max=0)
   n_elems = len(mesh['elems'])
   D = np.zeros((n_block,n_elems))
   np.add.at(D.T,mesh['i'],Gp.T)

   #Boundary----
   if len(mesh['db']) > 0: #boundary

      Gb = np.einsum('qj,jn->qn',sigma,mesh['db'],optimize=True)
      Gbp2 = Gb.clip(min=0); Gbm2 = Gb.clip(max=0);
      np.add.at(D.T,mesh['eb'],Gbp2[rr].T)
      tot = 1/Gb.clip(max=0).sum(axis=0); 
      GG  = np.einsum('qs,s->qs',Gbp2,tot)

   #Periodic---
   sss = np.asarray(mesh['pp'][:,0],dtype=int)
   P = np.zeros_like(D)
   np.add.at(P.T,mesh['i'][sss],-(mesh['pp'][:,1]*Gm[:,sss]).T)
   del G,Gp,Gbp2,tot
   #-------

   i = mesh['i']
   j = mesh['j']
   im = np.concatenate((mesh['i'],list(np.arange(n_elems))))
   jm = np.concatenate((mesh['j'],list(np.arange(n_elems))))
   data = np.concatenate((Gm,D+Wdiag[rr,None]),axis=1)
   X  = shared_array(np.tile(f['temperature'],(W.shape[0],1)))
   R  = shared_array(np.zeros_like(X))
   OP = shared_array(np.zeros_like(X))
   #--------------------

   #Winv = load_data('Winv')['Winv']


   kappa_mask = mesh['kappa_mask']/factor
   kappa_old  = f['meta'][0]
   eb = mesh['eb']

   (nm,n_elems) = X.shape

   #Post processing---

   def get_boundary(X):

      B = np.zeros((n_block,n_elems)) 
      #For now this is global--
      TB =  np.einsum('us,us->s',X[:,eb],GG)
      tmp = np.einsum('c,uc->cu',TB,Gbm2)
      #------------------------
      np.add.at(B.T,eb,tmp[:,rr])

      return B

   def L(X):

       OP[rr,:] = np.multiply(D,X[rr]) + sparse_dense_product(i,j,Gm,X[rr]) + np.matmul(W[rr,:],X) - get_boundary(X)

       comm.Barrier() 
       
       return OP


   def L_RTA(X):

       OP[rr,:] = np.multiply(D,X[rr]) + sparse_dense_product(i,j,Gm,X[rr]) + np.einsum('uc,u->uc',X,Wdiag)  - np.einsum('vc,v,u->uc',X,CT,Wdiag) - get_boundary(X)

       comm.Barrier() 

       return OP


   def residual_rta(X):
     R= L_RTA(X) 
     R[rr,:] -= P
     comm.Barrier()

     return -np.einsum('u,uc->uc',1/Wdiag,R)

   def residual(X):
     R = L(X) 
     R[rr,:] -= P
     comm.Barrier()
     return -R

   @cached(cache={})
   def get_lu(q):
        A = sp.csc_matrix((data[q],(im,jm)),shape=(n_elems,n_elems),dtype=np.float64)
        return sp.linalg.splu(A)


   def compute_kappa(X):
      kappa = np.zeros(1)
      kappap = np.einsum('u,uc,c->',sigma[rr,0],X[rr],kappa_mask)
      comm.Allreduce([np.array([kappap]),MPI.DOUBLE],[kappa,MPI.DOUBLE],op=MPI.SUM)
      comm.Barrier()
      return kappa[0]



   if comm.rank == 0:
    print(flush=True)
    print('      Iter    Thermal Conductivity [W/m/K]     Residual''',flush=True)
    print(colored(' -----------------------------------------------------------','green'),flush=True)

   alpha = argv.setdefault('alpha',1)  
   def solve_FULL(X,kappa_old):
    X_old = X.copy()   
    kk = 0
    error = 1
    error_old = 1
    r_old = 1e4
    while kk < max_bte_iter and r_old > tolerance:

      #B =  np.einsum('vc,v,u->uc',X_old,CT,Wdiag)[rr] + P + get_boundary(X_old)
      B        = -np.dot(Wod[rr],X_old) + P  + get_boundary(X_old)
      X[rr,:]  =  np.array([get_lu(n).solve(B[n]) for n,q in enumerate(rr)])
      
      kappa = compute_kappa(X)
      R     = residual(X)
      r = np.linalg.norm(R,ord='fro')
      if r > r_old and GGMRES and kk > 0:
         return X_old,kappa_old,True 
      r_old = r
      if comm.rank == 0:
       print('{0:8d} {1:24.4E} {2:22.4E}'.format(kk,kappa,r,flush=True))
      kk +=1
      comm.Barrier()
      X_old = alpha*X.copy()+ (1-alpha)*X_old


    return X,kappa,False
    
   X,kappa,gg = solve_FULL(X,kappa_old)

   if gg:
    X,kappa = ggmres(L,X,compute_kappa,residual,20,rr,max_ggmres_iter,tolerance)

   comm.Barrier()
   T = np.einsum('qc,q->c',X,CT)
   J = np.einsum('qj,qc->cj',sigma,X)*1e-18/factor

   bte =  {'kappa':[kappa],'temperature':T,'flux':J}
 
   argv['bte'] = bte


   









from setuptools import setup,find_packages
import os

setup(name='openbte',
      version='2.11',
      description='Boltzmann Transport Equation for Phonons',
      author='Giuseppe Romano',
      author_email='romanog@mit.edu',
      classifiers=['Programming Language :: Python :: 3.6'],
      #long_description=open('README.rst').read(),
      install_requires=['shapely',
                        'pyvtk',
                        'pandas',
                        'unittest2',
                        'nbsphinx',
                        'pytest',
                        'termcolor',
                        'alabaster',
                        'cachetools',
                        'mpi4py',
                        'plotly >=4.14.0',
                        'numpy',
                        'scipy',
                        'dash',
                        'dash_core_components',
                        'dash_html_components',
                        'dash_bootstrap_components',
                        'nbsphinx',
                        'recommonmark',
                        'sphinx',
                        'sphinx_rtd_theme'
                         ],
      license='GPLv3',\
      packages = ['openbte'],
      package_data = {'openbte':['materials/*.npz']},
      entry_points = {
     'console_scripts': ['AlmaBTE2OpenBTE=openbte.almabte2openbte:main',\
                         'Phono3py2OpenBTE=openbte.phono3py2openbte:main',\
                         'rta2mfp=openbte.rta2mfp:main',\
                         'bundle_data=openbte.bundle_data:main',\
                         'gui=openbte.gui:main',\
                         'app=openbte.app:App',\
                         'OpenBTE=openbte.openbte:main'],
      },
      zip_safe=False)



import sys
import numpy as np
import scipy
from mpi4py import MPI
import time
import os 
from openbte.utils import *


comm = MPI.COMM_WORLD

def get_linear_indexes(mfp,value):


   n = len(mfp)

   if value < mfp[0]:
      aj = (value-mfp[0])/(mfp[1]-mfp[0]) 
      return 0,1-aj,1,aj  
   elif value > mfp[-1]:
      aj = (value-mfp[n-2])/(mfp[n-1]-mfp[n-2]) 
      return n-2,1-aj,n-1,aj  
   else:
    for m in range(n-1):
      if (value <= mfp[m+1]) and (value >= mfp[m]) :
        aj = (value-mfp[m])/(mfp[m+1]-mfp[m]) 
        return m,1-aj,m+1,aj  
   



def mfp2DSym(**argv):

 #load data
 
 data = load_data('mfp')

 Kacc = data['Kacc']
 mfp_bulk  = data['mfp']

 kappa_bulk = Kacc[-1]
 kappa_bulk = np.zeros_like(Kacc)
 kappa_bulk[0] = Kacc[0]
 for n in range(len(Kacc)-1):
    kappa_bulk[n+1] = Kacc[n+1]-Kacc[n]

 #pruning--
 I = np.where(mfp_bulk > 0)
 kappa_bulk = kappa_bulk[I]
 mfp_bulk = mfp_bulk[I]
 kappa= np.eye(3) * np.sum(kappa_bulk)

 #Get options----
 n_phi = int(argv.setdefault('n_phi',48))
 n_mfp = int(argv.setdefault('n_mfp',50))
 n_theta = int(argv.setdefault('n_theta',24))

 nm = n_phi * n_mfp

 #Create sampled MFPs
 n_mfp_bulk = len(mfp_bulk)
 mfp_sampled = np.logspace(min([-2,np.log10(min(mfp_bulk)*0.99)]),np.log10(max(mfp_bulk)*1.01),n_mfp)#min MFP = 1e-2 nm 

 #Polar Angle---------
 Dphi = 2*np.pi/n_phi
 #phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 phi = np.linspace(0,2.0*np.pi,n_phi,endpoint=False)
 #--------------------
   
 #Azimuthal Angle------------------------------
 Dtheta = np.pi/n_theta/2.0
 theta = np.linspace(Dtheta/2.0,np.pi/2.0 - Dtheta/2.0,n_theta)
 dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)   
 domega = np.outer(dtheta,Dphi * np.ones(n_phi))

 #Compute directions---
 polar = np.array([np.sin(phi),np.cos(phi),np.zeros(n_phi)]).T
 azimuthal = np.array([np.sin(theta),np.sin(theta),np.zeros(n_theta)]).T
 direction = np.einsum('ij,kj->ikj',azimuthal,polar)
 
 #Compute average---
 ftheta = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
 fphi= np.sinc(Dphi/2.0/np.pi)
 polar_ave = polar*fphi
 azimuthal_ave = np.array([ftheta*np.sin(theta),ftheta*np.sin(theta),np.zeros(n_theta)]).T   
 direction_ave = np.einsum('ij,kj->ikj',azimuthal_ave,polar_ave)
 direction_int = np.einsum('ijl,ij->ijl',direction_ave,domega)
 #------------------------------------------------------

 n_mfp_bulk = len(mfp_bulk) 
 n_mfp = argv.setdefault('n_mfp',100)

 mfp = np.logspace(np.log10(max([min(mfp_bulk),1e-11])),np.log10(max(mfp_bulk)*1.01),n_mfp) 

 n_mfp = len(mfp)

 temp_coeff_p,temp_coeff = np.zeros((2,n_mfp,n_phi))
 kappa_directional_p,kappa_directional = np.zeros((2,n_mfp,n_phi,3)) 
 #suppression_p,suppression = np.zeros((2,n_mfp,n_phi,n_mfp_bulk)) 


 dirr = np.einsum('tpi,tp->tpi',direction_ave[:,:,:2],domega)

  
 kdp = np.zeros((n_mfp,n_phi,2))
 kd = np.zeros((n_mfp,n_phi,2))
 tcp = np.zeros((n_mfp,n_phi))
 tc = np.zeros((n_mfp,n_phi))
 p = np.arange(n_phi)


 g1 = kappa_bulk/mfp_bulk
 g2 = kappa_bulk/mfp_bulk/mfp_bulk
 g3 = ftheta*np.sin(theta)
 n_tot = n_mfp_bulk

 block =  n_tot//comm.size
 rr = range(block*comm.rank,n_tot) if comm.rank == comm.size-1 else range(block*comm.rank,block*(comm.rank+1))

 #change to mfp

 for t in range(n_theta):

   for m in rr:  
       
      (m1,a1,m2,a2) = get_linear_indexes(mfp,mfp_bulk[m]*g3[t])

      tmp = g1[m]*dirr[t]
      kdp[m1] += a1*tmp
      kdp[m2] += a2*tmp

      #Temperature
      tmp = g2[m]*domega[t]
      tcp[m1] += a1*tmp
      tcp[m2] += a2*tmp

      #Suppression
      #tmp  = dirr[t,:,0]/mfp_bulk[m]
      #sp[m1,:,m] +=  a1 * tmp
      #sp[m2,:,m] +=  a2 * tmp

   comm.Allreduce([kdp,MPI.DOUBLE],[kd,MPI.DOUBLE],op=MPI.SUM)
   comm.Allreduce([tcp,MPI.DOUBLE],[tc,MPI.DOUBLE],op=MPI.SUM)
   #comm.Allreduce([sp,MPI.DOUBLE],[s,MPI.DOUBLE],op=MPI.SUM)

 kd *= 2*3/4.0/np.pi
 #s *= 3*2/4.0/np.pi
 tc /= np.sum(tc)


 #Test for bulk---

 if comm.rank == 0:
  kappa_bulk = np.zeros((2,2))
  for m in range(n_mfp):
   for p in range(n_phi):
      kappa_bulk += mfp[m]*np.outer(kd[m,p],polar_ave[p,:2])


 #replicate bulk values---

 #Wod = np.tile(tc,(nm,1))

 #angle_map = np.arange(n_phi)
 #angle_map = np.repeat(angle_map,n_mfp)

 #repeat angle---
 #kappa_directional[:,:,2] = 0 #Enforce zeroflux on z (for visualization purposed)
 #F = np.einsum('m,pi->mpi',mfp,polar_ave)
 rhs_average = mfp_sampled*mfp_sampled/2
 a = np.zeros((3,3)) 
 for i in range(len(polar_ave)):
  a += np.outer(polar_ave[i],polar_ave[i])/2/np.pi*2*np.pi/n_phi
 rhs_average = mfp_sampled*mfp_sampled
 rhs_average *= a[0,0]
 
 #Final----
 return {'tc':tc,\
         'sigma':kd,\
         'kappa':kappa,\
         'mfp_average':rhs_average*1e18,\
         'VMFP':polar_ave[:,:2],\
         'mfp_sampled':mfp,\
         'sampling': np.array([n_phi,n_theta,n_mfp]),\
         'model':np.array([5]),\
         'suppression':np.zeros(1),\
         'kappam':kappa_bulk}





from mpi4py import MPI
from google_drive_downloader import GoogleDriveDownloader as gdd
import numpy as np
from termcolor import colored, cprint 
from shapely.geometry import Polygon,Point
from shapely.geometry import MultiPolygon,LineString
import shapely
from shapely.ops import cascaded_union
import os,sys
import functools
import io,zipfile
import math
import pickle
import gzip



def ggmres(L,X,compute_kappa,residual,k,rr,max_ggmres_iter,tolerance):

      def compute_norm_D(A,D):
       return np.sqrt(np.einsum('ij,i,ij->',A,D,A))

      if comm.rank == 0:
       print(' ')
       print(colored('                        G-GMRES','green'),flush=True)
       print(colored(' -----------------------------------------------------------','green'),flush=True)

      h = np.zeros(k+1)

      def arnoldi(Q,H):

       for j in range(k):
         #Arnoldi-------------------------------------
         V         = L(Q[j]) #shared
         hp         = np.einsum('ikj,k,kj->i',Q[:,rr,:],DD[rr],V[rr])
         comm.Allreduce([np.array([hp]),MPI.DOUBLE],[h,MPI.DOUBLE],op=MPI.SUM)

         H[:,j]    = h
         V[rr]    -= np.einsum('ikj,i->kj',Q[:,rr],H[:,j])
         comm.Barrier()
         H[j+1,j]  = compute_norm_D(V,DD)
         Q[j+1]    = V/H[j+1,j]   

       return Q,H
  
      r = 1e6
      kk = 0
      while kk < max_ggmres_iter and r > tolerance:

       X0 = X
       R = residual(X0) #shared

       #Compute WEIGHTS------------------------------
       DD = np.mean(np.absolute(R),axis=1)
       #---------------------------------------------  

       R0   = compute_norm_D(R,DD)
       Q = shared_array(np.zeros((k+1,X.shape[0],X.shape[1])))
       Q[0] = R/R0
       H = np.zeros((k+1,k))
       beta2 = np.zeros(k+1);beta2[0] = R0

       Q,H = arnoldi(Q,H)
       #--------------
       comm.Barrier() 
       c   = np.linalg.lstsq(H,beta2)[0]
       X   = shared_array(X0 + np.einsum('knm,k->nm',Q[:-1],c))
       r   = np.linalg.norm(residual(X),ord='fro')
       kappa = compute_kappa(X)
     
       if comm.rank == 0:
        print('{0:8d} {1:24.4E} {2:22.4E}'.format(kk,kappa,r),flush=True)

       kk +=1
       Q[:,:,:] = 0

      return X,kappa


def cg(L, b,**argv):

    #PARSE---------------------------------------
    x  = argv.setdefault('x0',np.zeros_like(b))  
    M  = argv.setdefault('M',lambda x: x.copy())  
    max_error  = argv.setdefault('error',1e-4)  
    callback  = argv.setdefault('callback',None) 
    verbose  = argv.setdefault('verbose',True) 
    
    #-------------------------------------------
    n = len(b)
    r = b - L(x)
    z = M(r)
    p = z
    r_k_norm = np.dot(r, z)
    r_0_norm = r_k_norm
    error = 1
    for i in range(2*n):
        Ap = L(p)
        alpha = r_k_norm / np.dot(p, Ap)
        x += alpha * p
        r -= alpha * Ap
        z = M(r)
        r_kplus1_norm = np.dot(r, z)

        error = r_kplus1_norm/r_0_norm
        if not callback == None:
            kappa = callback(x) 

        if verbose:
         print("Iter : % 4i, Kappa: % 5.3E  Error : % 5.2E" %(i, kappa,error))
        if error < max_error:
         break
        if i == 100:
            break

        beta = r_kplus1_norm / r_k_norm
        r_k_norm = r_kplus1_norm
        p = z + beta * p
    return x



def fix_instability(F,B,scale=True):

   n_elems = F.shape[0]
   if scale:
    scale = 1/F.max(axis=0).toarray()[0]
   else: 
    scale = np.ones(n_elems)   
   n = np.random.randint(n_elems)
   scale[n] = 0
   F.data = F.data * scale[F.indices]
   F[n,n] = 1
   B[n] = 0   

   return scale

def compute_kappa(W,b):


 L   = lambda x: np.dot(W,x)
 M   = lambda x: x/np.sqrt(np.diag(W))

 def callback(x):
     return np.dot(x,b)

 x0 = b.copy()/np.diag(W)
 x  = cg(L,b,M=M,x0=x0,callback=callback,verbose = False)

 return callback(x)


os.environ['H5PY_DEFAULT_READONLY']='1'

comm = MPI.COMM_WORLD

def fast_interpolation(fine,coarse,bound=False) :

 m2 = np.argmax(coarse >= fine[:,np.newaxis],axis=1)
 m1 = m2-1
 a2 = (fine-coarse[m1])/(coarse[m2]-coarse[m1])
 a1 = 1-a2

 if bound == 'periodic':
  Delta = coarse[-1]-coarse[-2]  
  a = np.where(m2==0)[0] #this can be either regions
  m1[a] = len(coarse) -1
  m2[a] = 0
  fine[fine < Delta/2] += 2*np.pi 
  a2[a] = (fine[a] - coarse[-1])/ Delta
  a1 = 1-a2


 if bound == 'extent':

   #Small values
   al = np.where(fine<coarse[0])[0] 
   m2[al] = 1; 
   m1[al] = 0;
   a2[al] = (fine[al]-coarse[0])/ (coarse[1]-coarse[0])
   a1[al] = 1-a2[al]


   #Large values
   ar = np.where(fine>coarse[-1])[0]
   m2[ar] = len(coarse)-1; 
   m1[ar] = len(coarse)-2;
   a2[ar] = (fine[ar]-coarse[-2])/ (coarse[-1]-coarse[-2])
   a1[ar] = 1-a2[ar]


 return a1,a2,m1,m2






def periodic_kernel(x1, x2, p,l,variance):
    return variance*np.exp(-2/l/l * np.sin(np.pi*abs(x1-x2)/p) ** 2)

def gram_matrix(xs,p,l,variance):
    return [[periodic_kernel(x1,x2,p,l,variance) for x2 in xs] for x1 in xs]

def generate_random_interface(p,l,variance,scale):

 xs = np.arange(-p/2, p/2,p/200)
 mean = [0 for x in xs]
 gram = gram_matrix(xs,p,l,variance)
 ys = np.random.multivariate_normal(mean, gram)*scale
 f = interpolate.interp1d(xs, ys,fill_value='extrapolate')

 return f


def make_polygon(Na,A):

 
   dphi = 2.0*math.pi/Na;
   r = math.sqrt(2.0*A/Na/math.sin(2.0 * math.pi/Na))
   poly_clip = []
   for ka in range(Na):
     ph =  dphi/2 + (ka-1) * dphi
     px  = r * math.cos(ph) 
     py  = r * math.sin(ph) 
     poly_clip.append([px,py])


   return poly_clip  


def create_loop(loops,line_list,store):

   #create external loop

   strc = 'Line Loop(' + str(loops)+ ') = {'
   for n in range(len(line_list)):
    strc +=str(line_list[n])
    if n == len(line_list)-1:
     strc += '};\n'
    else:
     strc += ','
   store.write(strc)

   return strc


def create_line_list(pp,points,lines,store,lx,ly,step_label = 'h'):

   def line_exists_ordered_old(l,lines):
    for n,line in enumerate(lines) :
     if (line[0] == l[0] and line[1] == l[1]) :
      return n+1
     if (line[0] == l[1] and line[1] == l[0]) :
      return -(n+1)
    return 0


   def already_included_old(all_points,new_point,p_list):

    dd = 1e-12
    I = []
    if len(all_points) > 0:
     I = np.where(np.linalg.norm(np.array(new_point)-np.array(all_points),axis=1)<dd)[0]
    if len(I) > 0: 
       return I[0]
    else: 
      return -1

   p_list = []
   for p in pp:
    tmp = already_included_old(points,p,p_list)
    if tmp == -1:    
      points.append(p)
      p_list.append(len(points)-1)
    else:
      p_list.append(tmp)

   for k,p in enumerate(p_list):
     point = points[-len(p_list)+k] 
     store.write( 'Point('+str(len(points)-len(p_list)+k) +') = {' + str(point[0]/lx) +'*lx,'+ str(point[1]/ly)+'*ly,0,' + step_label + '};\n')

   line_list = []
   for l in range(len(p_list)):
    p1 = p_list[l]
    p2 = p_list[(l+1)%len(p_list)]
    if not p1 == p2:
     tmp = line_exists_ordered_old([p1,p2],lines)

     if tmp == 0 : #craete line
      lines.append([p1,p2])
      store.write( 'Line('+str(len(lines)) +') = {' + str(p1) +','+ str(p2)+'};\n')
      line_list.append(len(lines))
     else:
      line_list.append(tmp)

   
   return line_list


 


def save_data(fh, namedict):

     if comm.rank == 0:
      with gzip.GzipFile(fh + '.npz', 'w') as f:
            pickle.dump(namedict, f,protocol=pickle.HIGHEST_PROTOCOL)

def load(fh):

    return  {f:load_data(f) for f in fh}


def load_data(fh):

    if comm.rank == 0:
     if os.path.isfile(fh + '.npz'):
      with gzip.open(fh + '.npz', 'rb') as f:
          return pickle.load(f)
     return -1 


def download_file(file_id,filename):

      gdd.download_file_from_google_drive(file_id=file_id,
                                           dest_path='./' + filename,showsize=True,overwrite=True)



def compute_neighbors(mesh,elem):
  
    guess = []

    for side in mesh['elem_side_map_vec'][elem]:
        for elem_2 in mesh['side_elem_map_vec'][side]:
            if not elem_2 in guess:
               guess.append(elem_2)
               
    return guess
                


def find_elem(mesh,p,guess):
 
    #First try in guess--
    for ne in guess:
     elem  = mesh['elems'][ne][0:mesh['side_per_elem'][ne]] 
     nodes = mesh['nodes'][elem][:,0:2] 
     polygon = Polygon(mesh['nodes'][elem][0:mesh['side_per_elem'][ne],0:2])
     if polygon.contains(Point(p[0],p[1])):
      return ne,nodes[:,0],nodes[:,1]
    #------------------      

    #Find among all elements
    for ne in range(len(mesh['elems'])):

     elem  = mesh['elems'][ne][0:mesh['side_per_elem'][ne]]  
     nodes = mesh['nodes'][elem][:,0:2]

     polygon = Polygon(nodes)
     if polygon.contains(Point(p[0],p[1])):
      return ne,nodes[:,0],nodes[:,1]

    return -1
        


def generate_frame(**argv):


    argv.setdefault('bounding_box',[0,0,0])
    Lx = float(argv.setdefault('lx',argv['bounding_box'][0]))
    Ly = float(argv.setdefault('ly',argv['bounding_box'][1]))
    frame = []
    frame.append([-Lx/2,Ly/2])
    frame.append([Lx/2,Ly/2])
    frame.append([Lx/2,-Ly/2])
    frame.append([-Lx/2,-Ly/2])

    return frame


def translate_shape(shape,base):

  out = []
  for p in shape:
    tmp = [p[0] + base[0],p[1] + base[1]]
    out.append(tmp)

  return out

def repeat_merge_scale(argv):

  polygons = argv['polygons']


  lx = argv['lx']   
  ly = argv['ly']   
  
  
  if argv.setdefault('relative',True):
    dx = 1;dy=1
  else:  
    dx = lx;dy=ly
  

  pbc = []
  pbc.append([0,0])
  if argv.setdefault('repeat',True):
   pbc.append([dx,0]) 
   pbc.append([-dx,0])
   pbc.append([0,dy])
   pbc.append([0,-dy])
   pbc.append([dx,dy])
   pbc.append([-dx,-dy])
   pbc.append([-dx,dy])
   pbc.append([dx,-dy])

  frame = Polygon([[-dx/2,-dy/2],[-dx/2,dy/2],[dx/2,dy/2],[dx/2,-dy/2]])

   #---------------------
  #Store only the intersecting polygons
  final = []
  for poly in polygons:
    for kp in range(len(pbc)):
     tmp = []
     for p in poly:
      cx = p[0] + pbc[kp][0]
      cy = p[1] + pbc[kp][1]
      tmp.append([cx,cy])
     p1 = Polygon(tmp)
     if p1.intersects(frame):
      thin = Polygon(p1).intersection(frame)
      if isinstance(thin, shapely.geometry.multipolygon.MultiPolygon):
       tmp = list(thin)
       for t in tmp:
        final.append(t)
      else:
        #final.append(thin)
        final.append(p1)

  #print(list(final[0].exterior.coords))
  #quit()
  #Create bulk surface---get only the exterior to avoid holes
  MP = MultiPolygon(final) 

  conso = cascaded_union(MP)

  polygons_final = []
  if isinstance(conso, shapely.geometry.multipolygon.MultiPolygon):
      for i in conso: 
       #polygons_final.append(list(i.simplify(1e-2).exterior.coords))
       polygons_final.append(list(i.exterior.coords))
  else: 
       polygons_final.append(list(conso.exterior.coords))
  

  #cut redundant points
  new_poly = []
  for poly in polygons_final:
   N = len(poly)
   tmp = []
   for n in range(N):
    p1 = poly[n]
    p2 = poly[(n+1)%N]
    if np.linalg.norm(np.array(p1)-np.array(p2)) >1e-4:
     tmp.append(p1)
   new_poly.append(tmp)

  #------------------------------------------

  #cut points which are in line--

  #cut redundant points
  


  #p = Polygon(polygons_final[0]).simplify(1e-2)
  #print(len(p.exterior.coords))
  #new_poly = polygons_final.copy()
  discard = 1
  while discard > 0:
      
   discard = 0  
   new_poly_2 = []
   for gg,poly in enumerate(new_poly):
    N = len(poly)
    tmp = []
    for n in range(N):
     p1 = np.array(poly[(n-1+N)%N])
     p = np.array(poly[n])
     p2 = np.array(poly[(n+1)%N])
     di = np.linalg.norm(np.cross(p-p1,p-p2))

     if di >1e-5:
        tmp.append(p)
     else:
        discard += 1 
    new_poly_2.append(tmp)
   new_poly = new_poly_2.copy()


  #----------------------------
  dmin = check_distances(new_poly)


  #scale-----------------------
  if argv.setdefault('relative',True):
   polygons = []
   for poly in new_poly:
    tmp = []
    for p in poly:
     g = list(p)
     g[0] *=lx
     g[1] *=ly
     tmp.append(g)
    polygons.append(tmp) 

  argv['polygons'] = polygons
  argv['dmin'] = dmin


def check_distances(new_poly):


  #Merge pores is they are too close
  Pores = [ Polygon(p)  for p in new_poly]
  dmin = 1e4
  move = {}
  for p1 in range(len(Pores)):
   #check if intersect  
   
   for p2 in range(p1+1,len(Pores)):
       d = Pores[p1].distance(Pores[p2])
       if d > 0 and d < dmin : dmin = d
      
   #top
   d = Pores[p1].distance(LineString([(-0.5,0.5),(0.5,0.5)])) 
   if d < dmin and d > 0:
     dmin = d

   #left
   d = Pores[p1].distance(LineString([(0.5,0.5),(0.5,-0.5)])) 
   if d < dmin and d > 0:
     dmin = d

   #bottom
   d = Pores[p1].distance(LineString([(-0.5,-0.5),(0.5,-0.5)])) 
   if d < dmin and d > 0:
     dmin = d
  
   #left
   d = Pores[p1].distance(LineString([(-0.5,-0.5),(-0.5,0.5)])) 
   if d < dmin and d > 0:
     dmin = d

  return dmin



def interpolate(vector,value,bounds='extent',period = None):

   n = len(vector)

   if value >= vector[0] and value <= vector[-1]:
     
    for m in range(n-1):
      if (value <= vector[m+1]) and (value >= vector[m]) :
        i = m; j = m + 1 
        break
    aj = (value-vector[i])/(vector[j]-vector[i]) 
    ai = 1-aj
    return i,ai,j,aj  

   else:

    if bounds == 'extent': 
     if value < vector[0]:
       i=0;j=1;
     elif value > vector[-1]:
       i=n-2;j=n-1;
     aj = (value-vector[i])/(vector[j]-vector[i]) 
     ai = 1-aj
     return i,ai,j,aj 

    elif bounds == 'periodic':     
     i=n-1;j=0;
     if value < vector[0]:
       aj = (value + period -vector[i])/(vector[j] + period -vector[i])
     elif value > vector[-1]:
       aj = (value - vector[-1])/(vector[0] + period - vector[-1]) 

     ai = 1-aj
     return i,ai,j,aj 




def get_linear_indexes(mfp,value,scale,extent):

   if value == 0:
     return -1,-1,-1,-1

   n = len(mfp)

   if scale == 'log':
     mfp = np.log10(mfp.copy())
     value = np.log10(value.copy())

   found = False
   beyond = False
   if extent:
    if value < mfp[0]:
      beyond = True
      found = True
      i=0;j=1;
    elif value > mfp[-1]:
      i=n-2;j=n-1;
      beyond = True
      found = True

   if beyond == False:
    for m in range(n-1):
      if (value <= mfp[m+1]) and (value >= mfp[m]) :
        i = m; j = m + 1 
        found = True
        break
  
   if found == False:
      print('no interpolation found')
   else:  
    aj = (value-mfp[i])/(mfp[j]-mfp[i]) 
    ai = 1-aj
    if scale=='inverse':
     ai *=mfp[i]/value
     aj *=mfp[j]/value
   return i,ai,j,aj  


def shared_array(value):

    data = {'dummy':value}
    return create_shared_memory_dict(data)['dummy']


def get_kappa_map_from_mat(**argv):

        mat_map = argv['geometry']['elem_mat_map']
        kappa =   argv['material']['kappa']

        kappa = np.array(kappa)
        if kappa.ndim == 3:
         return np.array([ list(kappa[i])  for i in mat_map])
        else:
         return np.array([ list(kappa)  for i in mat_map])
            


def create_shared_memory_dict(varss):

       dict_output = {}
       if comm.Get_rank() == 0:
          var_meta = {} 
          for var,value in varss.items():

           if type(value) == list: value = np.array(value)   
           if value.dtype == np.int64:
              data_type = 0
              itemsize = MPI.INT.Get_size()
           elif value.dtype == np.float64:
              data_type = 1
              itemsize = MPI.DOUBLE.Get_size() 
           else:
              print('data type for shared memory not supported for ' + var)
              quit()

           size = np.prod(value.shape)
           nbytes = size * itemsize
           var_meta[var] = [value.shape,data_type,itemsize,nbytes]

       else: nbytes = 0; var_meta = None
       var_meta = comm.bcast(var_meta,root=0)

       #ALLOCATING MEMORY---------------
       for n,(var,meta) in enumerate(var_meta.items()):
       
        win = MPI.Win.Allocate_shared(meta[3],meta[2], comm=comm) 
        buf,itemsize = win.Shared_query(0)
        assert itemsize == meta[2]
        dt = 'i' if meta[1] == 0 else 'd'
        output = np.ndarray(buffer=buf,dtype=dt,shape=meta[0]) 
        if comm.rank == 0: output[:] = varss[var]
        dict_output[var] = output

       del varss
       comm.Barrier()

       return dict_output


def repeat_nodes_and_data(data_uc,nodes_uc,increment,repeat,size,cells_uc):

   Nx = repeat[0] 
   Ny = repeat[1] 
   Nz = repeat[2] 

   #----------------------------------------------
   n_nodes = len(nodes_uc)

   nodes = []
   data = []
   #add nodes---
   corr = np.zeros(n_nodes*Nx*Ny*Nz,dtype=np.int)
   index = 0
   for nx in range(Nx):
    for ny in range(Ny):
     for nz in range(Nz):
      dd = nx*increment[0] + ny*increment[1] + nz*increment[2]
      index = nx * Ny * Nz * n_nodes +  ny *  Nz * n_nodes + nz * n_nodes
      P = [round(size[0]*nx,3),round(size[1]*ny,3),round(size[2]*nz,3)]
      for n in range(n_nodes):
       node_trial = [round(nodes_uc[n,0],4) + P[0]-(Nx-1)*0.5*size[0],\
                     round(nodes_uc[n,1],4) + P[1]-(Nx-1)*0.5*size[1],\
                     round(nodes_uc[n,2],4) + P[2]-(Nx-1)*0.5*size[2]]

       nodes.append(node_trial)
       data.append(data_uc[n]-dd)
       corr[index+n] = len(nodes)-1

   cells = []
   for nx in range(Nx):
    for ny in range(Ny):
     for nz in range(Nz):
      P = [size[0]*nx,size[1]*ny,size[2]*nz]
      index = nx * Ny * Nz * n_nodes +  ny *  Nz * n_nodes + nz * n_nodes
      #add cells
      for gg in cells_uc:
       tmp = []
       for n in gg:
        tmp.append(corr[n+index])
       cells.append(tmp)

   return np.array(nodes),np.array(cells),np.array(data)
        

def create_shared_memory(varss):
       for var in varss:
         #--------------------------------------
         if comm.Get_rank() == 0: 
          tmp = eval('self.' + var)
          if tmp.dtype == np.int64:
              data_type = 0
              itemsize = MPI.INT.Get_size() 
          elif tmp.dtype == np.float64:
              data_type = 1
              itemsize = MPI.DOUBLE.Get_size() 
          else:
              print('data type for shared memory not supported')
              quit()
          size = np.prod(tmp.shape)
          nbytes = size * itemsize
          meta = [tmp.shape,data_type,itemsize]
         else: nbytes = 0; meta = None
         meta = comm.bcast(meta,root=0)

         #ALLOCATING MEMORY---------------
         win = MPI.Win.Allocate_shared(nbytes,meta[2], comm=comm) 
         buf,itemsize = win.Shared_query(0)
         assert itemsize == meta[2]
         dt = 'i' if meta[1] == 0 else 'd'
         output = np.ndarray(buffer=buf,dtype=dt,shape=meta[0]) 

         if comm.rank == 0:
             output[:] = tmp  

         exec('self.' + var + '=output')





from __future__ import absolute_import
import numpy as np
from scipy.sparse.linalg import splu
from termcolor import colored, cprint 
from .utils import *
from mpi4py import MPI
import sys
import scipy.sparse as sp
import time
from scipy.sparse.linalg import lgmres
import sys
from shapely.geometry import LineString
import profile
import scipy
from .solve_rta import *
from .solve_full import *
import pkg_resources  
from .fourier import *
import warnings

comm = MPI.COMM_WORLD

def get_model(m):

    models = ['Fourier','gray2D','gray2DSym','gray3D','mfp2D','mfp2DSym','mfp3D','rta2D','rta2DSym','rta3D','full','mg2DSym']


    return models[m[0]]

def fourier_info(argv):
          print('                        FOURIER                 ',flush=True)   
          print(colored(' -----------------------------------------------------------','green'),flush=True)
          #print(colored('  Iterations:                              ','green') + str(int(data[2])),flush=True)
          #print(colored('  Relative error:                          ','green') + '%.1E' % (data[1]),flush=True)
          #print(colored('  Fourier Thermal Conductivity [W/m/K]:    ','green') + str(round(data[0],3)),flush=True)
          print(colored(' -----------------------------------------------------------','green'),flush=True)
          print(" ")



def print_bulk_info(mat,mesh):

           print(colored('  Model:                                   ','green')+ get_model(mat['model']),flush=True)
          
           if mat['kappa'].ndim == 3:
              for kappa in mat['kappa']:    
                print(colored('  Bulk Thermal Conductivity (xx) [W/m/K]:  ','green')+ str(round(kappa[0,0],2)),flush=True)
                print(colored('  Bulk Thermal Conductivity (yy) [W/m/K]:  ','green')+ str(round(kappa[1,1],2)),flush=True)
                if self.dim == 3:
                  print(colored('  Bulk Thermal Conductivity (zz)[W/m/K]:   ','green')+ str(round(kappa[2,2],2)),flush=True)
           else: 
                kappa = mat['kappa']
                print(colored('  Bulk Thermal Conductivity (xx) [W/m/K]:  ','green')+ str(round(kappa[0,0],2)),flush=True)
                print(colored('  Bulk Thermal Conductivity (yy) [W/m/K]:  ','green')+ str(round(kappa[1,1],2)),flush=True)
                if len(kappa) == 3:
                  print(colored('  Bulk Thermal Conductivity (zz)[W/m/K]:   ','green')+ str(round(kappa[2,2],2)),flush=True)

           dirr = ['x','y','z']
           print(colored('  Applied Thermal Gradient along :         ','green')+ dirr[int(mesh['meta'][-1])],flush=True)


def print_options(**argv):
          print(colored('  Multiscale:                              ','green')+ str(argv['multiscale']),flush=True)
          print(colored('  Relaxation:                              ','green')+ str(argv['alpha']),flush=True)
          print(colored('  Keep Lu:                                 ','green')+ str(argv['keep_lu']),flush=True)
          print(colored('  Load State                               ','green')+ str(argv['load_state']),flush=True)
          print(colored('  Save State                               ','green')+ str(argv['save_state']),flush=True)
          print(colored('  Multiscale Error (Fourier):              ','green')+ str(argv['multiscale_error_fourier']),flush=True)
          print(colored('  Multiscale Error (Ballistic):            ','green')+ str(argv['multiscale_error_ballistic']),flush=True)
          print(colored('  Only Fourier:                            ','green')+ str(argv['only_fourier']),flush=True)
          print(colored('  Max Fourier Error:                       ','green')+ '%.1E' % (argv['max_fourier_error']),flush=True)
          print(colored('  Max Fourier Iter:                        ','green')+ str(argv['max_fourier_iter']),flush=True)
          print(colored('  Max BTE Error:                           ','green')+ '%.1E' % (argv['max_bte_error']),flush=True)
          print(colored('  Max BTE Iter:                            ','green')+ str(argv['max_bte_iter']),flush=True)


def print_logo():

    v = pkg_resources.require("OpenBTE")[0].version   
    print(' ',flush=True)
    print(colored(r'''        ___                   ____ _____ _____ ''','green'),flush=True)
    print(colored(r'''       / _ \ _ __   ___ _ __ | __ )_   _| ____|''','green'),flush=True)
    print(colored(r'''      | | | | '_ \ / _ \ '_ \|  _ \ | | |  _|  ''','green'),flush=True)
    print(colored(r'''      | |_| | |_) |  __/ | | | |_) || | | |___ ''','green'),flush=True)
    print(colored(r'''       \___/| .__/ \___|_| |_|____/ |_| |_____|''','green'),flush=True)
    print(colored(r'''            |_|                                ''','green'),flush=True)
    print(colored(r'''                                               v. ''' + str(v) ,'blue'),flush=True)
    print(flush=True)
    print('                       GENERAL INFO',flush=True)
    print(colored(' -----------------------------------------------------------','green'),flush=True)
    print(colored('  Contact:          ','green') + 'romanog@mit.edu                        ',flush=True) 
    print(colored('  Source code:      ','green') + 'https://github.com/romanodev/OpenBTE     ',flush=True)
    print(colored('  Become a sponsor: ','green') + 'https://github.com/sponsors/romanodev    ',flush=True)
    print(colored('  Documentation:    ','green') + 'https://openbte.readthedocs.io',flush=True)
    print(colored(' -----------------------------------------------------------','green'),flush=True)
    print(flush=True)   


def prepare_data(argv):
        
  #Original Description--
  tmp = {0:{'name':'Temperature Fourier','units':'K','data':argv['fourier']['temperature'],'increment':-argv['geometry']['applied_gradient']},\
               1:{'name':'Flux Fourier'       ,'units':'W/m/m','data':argv['fourier']['flux'],'increment':[0,0,0]}}

  if not argv.setdefault('only_fourier',False):
      tmp[2]    = {'name':'Temperature BTE','units':'K','data':argv['bte']['temperature'],'increment':-argv['geometry']['applied_gradient']}
      tmp[3]    = {'name':'Flux BTE'       ,'units':'W/m/m','data':argv['bte']['flux'],'increment':[0,0,0]}
              
  #Here we unroll variables for later use--
  variables = {}
  for key,value in tmp.items():
      
     #if value['data'].ndim == 1: #scalar
     variables[value['name']] = {'data':value['data'],'units':value['units'],'increment':value['increment']}
     #elif value['data'].ndim == 2 : #vector 
     #    variables[value['name'] + '(x)'] = {'data':value['data'][:,0],'units':value['units'],'increment':value['increment']}
     #    variables[value['name'] + '(y)'] = {'data':value['data'][:,1],'units':value['units'],'increment':value['increment']}
     #    if argv['dim'] == 3: 
     #        variables[value['name'] + '(z)'] = {'data':value['data'][:,2],'units':value['units'],'increment':value['increment']}
     #    mag = np.array([np.linalg.norm(value) for value in value['data']])
     #    variables[value['name'] + '(mag.)'] = {'data':mag,'units':value['units'],'increment':value['increment']}
  argv['variables'] = variables

  dirr = int(argv['geometry']['meta'][-1])
  argv['data'] = {'variables':variables,'kappa_bulk':argv['material']['kappa'][dirr,dirr] ,'kappa_fourier':argv['fourier']['meta'][0]}

  if 'bte' in argv.keys():
     argv['data']['kappa_bte']  = argv['bte']['kappa'][-1]  
     if 'kappa_mode' in argv['bte']:
       argv['data']['kappa_mode'] = argv['bte']['kappa_mode']
     if 'kappa_mode_f' in argv['bte']:
       argv['data']['kappa_mode_f'] = argv['bte']['kappa_mode_f']
     if 'kappa_0' in argv['bte']:
       argv['data']['kappa_0'] = argv['bte']['kappa_0']




def Solver(**argv):


        #COMMON OPTIONS------------
        argv.setdefault('multiscale',False)
        argv.setdefault('multiscale_error_fourier',5e-2)
        argv.setdefault('multiscale_error_ballistic',1e-2)
        argv.setdefault('verbose',True)
        argv.setdefault('save_state',False)
        argv.setdefault('load_state',False)
        argv.setdefault('alpha',1.0)
        argv.setdefault('keep_lu',True)
        argv.setdefault('only_fourier',False)
        argv.setdefault('max_bte_iter',20)
        argv.setdefault('max_bte_error',1e-3)
        argv.setdefault('max_fourier_iter',20)
        argv.setdefault('deviational',False)
        argv.setdefault('max_fourier_error',1e-5)
        #----------------------------
       

        mesh = None 
        mat = None 
        if comm.rank == 0 :
            if argv['verbose']:
             print_logo()
             print('                         SYSTEM                 ',flush=True)   
             print(colored(' -----------------------------------------------------------','green'),flush=True)
             print_options(**argv)

            mesh = argv['geometry'] if 'geometry' in argv.keys() else load_data(argv.setdefault('geometry_file','geometry'))
            mat  = argv['material'] if 'material' in argv.keys() else load_data(argv.setdefault('material_file','material'))
            
            if argv['verbose']:
             print_bulk_info(mat,mesh)
             print_mpi_info()
             print_grid_info(mesh,mat)

        if comm.size > 1:

         #load geometry---
         argv['geometry'] = create_shared_memory_dict(mesh)
 
         #load material---
         argv['material'] = create_shared_memory_dict(mat)

         argv['geometry']['elem_kappa_map'] = get_kappa_map_from_mat(**argv)   

         #Solve fourier--
         argv['fourier'] = create_shared_memory_dict(solve_fourier_single(argv))

        else:

         #load geometry---
         argv['geometry'] = mesh
 
         #load material---
         argv['material'] = mat

         argv['geometry']['elem_kappa_map'] = get_kappa_map_from_mat(**argv)

         #Solve fourier--
         argv['fourier'] = solve_fourier_single(argv)


        #Solve bte--
        if not argv['only_fourier']:

           mat_model = get_model(argv['material']['model'])

           if mat_model in ['rta2DSym','rta2D','rta3D','mfp2DSym','mfp2D','gray2D','gray2DSym','mg2DSym'] :
               solve_rta(argv)

           elif mat_model == 'full':    
               solve_full(argv)

           elif mat_model == 'exp':    
               solve_exp(argv)

           #elif mat_model in ['mg2DSym']:    
           #    solve_mg(argv)

           else:
               print('No model coded')
               quit()

        
        argv['dim'] =  int(argv['geometry']['meta'][2])
        prepare_data(argv)
        if comm.rank == 0 and argv.setdefault('save',True):
           save_data(argv.setdefault('filename','solver'),argv['data'])   

        #Clear cache
        clear_fourier_cache() 
        clear_BTE_cache() 
        if argv['verbose'] and comm.rank == 0:
         print(' ',flush=True)   
         print(colored('                 OpenBTE ended successfully','green'),flush=True)
         print(' ',flush=True)  


        return argv['data']


def print_mpi_info():

          print(colored('  vCPUs:                                   ','green') + str(comm.size),flush=True)
          print(" ")



def print_grid_info(mesh,mat):

          dim     = int(mesh['meta'][2])
          n_elems = len(mesh['elems'])
          print('                          GRID                 ')   
          print(colored(' -----------------------------------------------------------','green'))
          print(colored('  Dimension:                               ','green') + str(dim),flush=True)
          print(colored('  Size along X [nm]:                       ','green')+ str(round(mesh['size'][0],2)),flush=True)
          print(colored('  Size along y [nm]:                       ','green')+ str(round(mesh['size'][1],2)),flush=True)
          if dim == 3:
           print(colored('  Size along z [nm]:                       ','green')+ str(round(mesh['size'][2],2)),flush=True)
          print(colored('  Number of Elements:                      ','green') + str(n_elems),flush=True)
          print(colored('  Number of Sides:                         ','green') + str(len(mesh['active_sides'])),flush=True)
          print(colored('  Number of Nodes:                         ','green') + str(len(mesh['nodes'])),flush=True)


          with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            if mat['model'] in [8,9]:   
             print(colored('  Number of MFP:                           ','green')+ str(mat['tc'].shape[0]),flush=True)
             print(colored('  Number of Solid Angles:                  ','green')+ str(mat['tc'].shape[1]),flush=True)
            else : 
             print(colored('  Number of wave vectors:                  ','green')+ str(mat['sigma'].shape[0]),flush=True)

          if dim == 3:
           filling = np.sum(mesh['volumes'])/mesh['size'][0]/mesh['size'][1]/mesh['size'][2]
          else: 
           filling = np.sum(mesh['volumes'])/mesh['size'][0]/mesh['size'][1]
          print(colored('  Computed porosity:                       ','green') + str(round(1-filling,3)),flush=True)
          
          print(colored(' -----------------------------------------------------------','green'))
          print(" ")






